// Reference solution for PE2 Spring 2022.
// Much of the code below was provided with the exam.
// Functions to be authored by students are marked "SOLUTION CODE:"

public class MovieMatcher {

    private ST<String, Preferences> people; // map from person name to their Preferences
    private String[] movies; // array of movie titles that have been rated

    // Constructor: assign the instance variables (creating two empty symbol tables).
    public MovieMatcher() {
        people = new ST<String, Preferences>(); 
        movies = null; // No movies have been rated yet. This array gets filled in below.
    }

    // Read ratings from standard in, until empty.
    // Each line has a person, a rating, and movie, like this:
    //   Angel -1 Gone_with_the_Wind
    // The rating is +1 for "thumbs up" or -1 for "thumbs down".
    public void addRatingsFromStdIn() {
        SET<String> titles = new SET<String>();

        while(!StdIn.isEmpty()) {
            // read one line from StdIn
            String person = StdIn.readString();
            int    rating = StdIn.readInt();
            String movie  = StdIn.readString();

            titles.add(movie);

            // get (or create) the preferences for this person
            Preferences prefs = people.get(person);
            if (prefs == null) {
                prefs = new Preferences(person);
                people.put(person, prefs);
            }
            // save this movie rating in this person's preferences
            prefs.addRating(movie, rating);
        }
        // save all movie titles into our movies array (from set)
        int n = titles.size();
        movies = new String[n];
        n = 0;
        for (String title : titles) {
            movies[n] = title;
            n++;
        }
    }

    // Return the preferences for a person with the given name.
    public Preferences getPersonPreferences(String person) {
        // SOLUTION CODE:
        return people.get(person);
    }

    // Which other person is most similar to given person?
    // Return their name, or empty string if there is nobody else.
    public String mostSimilar(String person) {
        // SOLUTION CODE:
        Preferences prefs = getPersonPreferences(person);
        String match = "nobody";
        int mostSim = Integer.MIN_VALUE;
        for (String other : people.keys()) {
            if (!other.equals(person)) {
                Preferences otherPrefs = people.get(other);
                int sim = prefs.similarity(otherPrefs);
                if (mostSim < sim) {
                    mostSim = sim;
                    match = other;
                }
            }
        }
        return match;
    }

    // Predict how a person would rate a movie based on prefs from other people.
    // Sum over all other people who rated this movie. If the other person is 
    // similar, add their rating (otherwise subtract it), weighted by how similar.
    // Return positive or negative number whose magnitude reflects confidence.
    public int predict(String person, String movie) {
        // SOLUTION CODE:
        Preferences prefs = getPersonPreferences(person);
        int prediction = 0;
        for (String other : people.keys()) {
            if (!other.equals(person)) {
                Preferences otherPrefs = people.get(other);
                if (otherPrefs.hasRating(movie)) {
                    int otherRating = otherPrefs.getRating(movie);
                    int sim = prefs.similarity(otherPrefs);
                    prediction += sim * otherRating;
                }
            }
        }
        return prediction;
    }

    // Which unseen movie has the highest prediction for this person?
    public String recommend(String person) {
        // SOLUTION CODE:
        Preferences prefs = getPersonPreferences(person);
        String match = "nothing";
        int bestPred = Integer.MIN_VALUE;
        for (int i = 0; i < movies.length; i++) {
            String movie = movies[i];
            if (!prefs.hasRating(movie)) {
                int pred = predict(person, movie);
                if (bestPred < pred) {
                    bestPred = pred;
                    match = movie;
                }
            }
        }
        return match;
    }

    // Return string representation of this object: each person, with their ratings.
    public String toString() {
        String result = "";
        for (String person : people.keys()) {
            Preferences prefs = getPersonPreferences(person);
            result += prefs;
        }
        return result;
    }

    public static void printUsage() {
        String message = "Options for running this program:\n\n"
        + "java-introcs MovieMatcher\n" 
        + " --- print this message\n\n" 
        + "java-introcs MovieMatcher OUTPUT < ratings.txt\n" 
        + " --- read in movie preferences and print out MovieMatcher\n\n"
        + "java-introcs MovieMatcher SIMILAR Angel < ratings.txt\n" 
        + " --- read in movie preferences and output which person is most similar to Angel\n\n"
        + "java-introcs MovieMatcher PREDICT Angel Star_Wars < ratings.txt\n" 
        + " --- read movie preferences and predict how strongly Angel would like Star_Wars\n\n"
        + "java-introcs MovieMatcher RECOMMEND Angel < ratings.txt\n" 
        + " --- read movie preferences and recommend the movie Angel is most likely to enjoy\n";
        StdOut.println(message);
    }

    public static void main(String[] args) {
        if (args.length >= 1 && args[0].equals("OUTPUT")) {
            MovieMatcher mm = new MovieMatcher();
            mm.addRatingsFromStdIn();
            StdOut.println(mm);
        }
        else if (args.length >= 2 && args[0].equals("SIMILAR")) {
            MovieMatcher mm = new MovieMatcher();
            mm.addRatingsFromStdIn();
            String person  = args[1];
            String similar = mm.mostSimilar(person);
            StdOut.println("Most similar to " + person + " is: " + similar);
        }
        else if (args.length >= 3 && args[0].equals("PREDICT")) {
            MovieMatcher mm = new MovieMatcher();
            mm.addRatingsFromStdIn();
            String person = args[1];
            String movie  = args[2];
            int pred = mm.predict(person, movie);
            StdOut.println("Prediction for " + person + " and " + movie + ": " + pred);
        }
        else if (args.length >= 2 && args[0].equals("RECOMMEND")) {
            MovieMatcher mm = new MovieMatcher();
            mm.addRatingsFromStdIn();
            String person = args[1];
            String rec = mm.recommend(person);
            StdOut.println("Recommendation for " + person + " is: " + rec);
        }
        else {
            printUsage();
        }
    }
}
