// Reference solution for PE2 Spring 2022.
public class Preferences {

    private String name; // the person's name
    private ST<String, Integer> st; // map from movie title to a rating (+1 or -1)

    // Constructor. Preferences for a person with this name.
    public Preferences(String name) {
        st = new ST<String, Integer>();
        this.name = name;
    }

    // Return the person's name.
    public String getName() {
        return name;
    }

    // Insert this person's rating for a particular movie.
    public void addRating(String movie, int rating) {
        st.put(movie, rating);
    }

    // Has this person rated this movie?
    public boolean hasRating(String movie) {
        return st.contains(movie);
    }

    // If this person has rated this movie, return the rating.
    // Otherwise, throw a custom runtime exception.
    public int getRating(String movie) {
        if (hasRating(movie)) {
            return st.get(movie);
        }
        String message = name + " did not rate " + movie;
        throw new RuntimeException(message);
    }

    // Return a character symbol indicating a given movie rating:
    // plus sign ('+') for +1 or lowercase 'x' for -1.
    public static char ratingSymbol(int rating) {
        if (rating > 0) {
            return '+';
        }
        return 'x';
    }

    // Return a string containing all movie ratings from this person.
    // See exam for details.
    public String toString() {
        String output = name + ":\n";
        for (String movie : st.keys()) {
            int rating = getRating(movie);
            char symbol = ratingSymbol(rating);
            output += symbol + " " + movie + "\n";
        }
        return output;
    }

    // Return similarity between this person's ratings and those of another person.
    // Defined as: only considering movies that both people rated,
    // the number of movies where they agree minus the number where they disagree.
    // Assume ratings are either +1 or -1.
    public int similarity(Preferences other) {
        int sim = 0;
        for (String movie : st.keys()) {
            if (other.hasRating(movie)) {
                int myRating = this.getRating(movie);
                int otherRating = other.getRating(movie);
                sim += myRating * otherRating;
            }
        }
        return sim;
    }

    // You can use the main method to test your functions above.
    // It has a default set of tests, but feel free to edit it as you prefer.
    public static void main(String[] args) {
        Preferences alice = new Preferences("Alice");
        Preferences bobby = new Preferences("Bobby");
        String movie1 = "Casablanca";
        String movie2 = "Star_Wars";
        String movie3 = "The_Graduate";
        StdOut.println("Alice, before adding ratings:");
        StdOut.println(alice);
        alice.addRating(movie1, +1);
        alice.addRating(movie2, -1);
        alice.addRating(movie3, +1);
        StdOut.println("Alice, after adding ratings:");
        StdOut.println(alice);
        bobby.addRating(movie1, +1);
        bobby.addRating(movie2, -1);
        StdOut.println("Bobby, after adding ratings:");
        StdOut.println(bobby);
        int sim = alice.similarity(bobby);
        StdOut.println("Alice-Bob similarity: " + sim);
        StdOut.println("This next function should throw an exception:");
        int err = bobby.getRating(movie3);
    }
}
